import os
from os.path import join, dirname, abspath
from typing import List

from hiword import extract_keywords


def _load_docs() -> List[str]:
    docs = []
    base_dir = join(dirname(abspath(__file__)), 'docs')
    for i in os.listdir(base_dir):
        if i.endswith('.txt'):
            with open(join(base_dir, i), 'r') as f:
                docs.append(f.read())
    return docs


def test_extract_keywords():
    docs = _load_docs()
    for i, s in enumerate(docs):
        print('#{}: {}'.format(i + 1, s))
        print('#keywords:')
        for x, w in extract_keywords(s, with_weight=True):
            print(x, w)



import math
from collections import defaultdict
from typing import Dict, List, Union

from LAC import LAC

from hiword.dataloader import DictLoader, IDFLoader, StopwordsLoader
from hiword.filter import WordFilterChain, NumericFilter
from hiword.utils import traditional_to_simple, filter_invalid_chars


class KeywordsExtractor:
    def __init__(self):
        self.lac = LAC(mode='seg')
        self.dict = DictLoader()
        self.idf = IDFLoader()
        self.stopwords = StopwordsLoader()

        self.word_filter = WordFilterChain(
            NumericFilter(),
        )

    def extract_keywords(self, doc):
        words = self._tokenize(doc)
        keywords = self._extract_short_keywords(words)
        long_words = self._detect_long_keywords(words, keywords)

        freq = defaultdict(lambda: 0)
        for word, count in keywords:
            if self.word_filter.filter(word):
                continue
            freq[word] += count
        for word, count in long_words:
            if self.word_filter.filter(word):
                continue
            freq[word] += count

        res = self._rerank_words(freq)
        res = self._filter_subwords(res, freq)

        # FIXME: 根据经验设置的权重阈值
        min_weight = 0.1
        res = [[k, v] for k, v in res if v >= min_weight]
        return res

    def _tokenize(self, doc: Union[str, List[str]]) -> List[str]:
        if isinstance(doc, str):
            words = []
            for i in self.lac.run(doc):
                w = i.strip()
                if len(w) == 0:
                    continue
                words.append(w)
        else:
            words = doc
        return words

    def _filter_subwords(self, words: list, freq: Dict[str, int]):
        merged = []
        for i in range(len(words)):
            w1 = words[i][0]
            w1_ok = True
            for j in range(len(words)):
                w2 = words[j][0]
                if len(w1) < len(w2) and w1 in w2 and freq[w1] == freq[w2]:
                    w1_ok = False
                    break
            if w1_ok:
                merged.append((words[i][0], words[i][1]))
        return merged

    def _detect_long_keywords(self, words, short_words):
        keywords = set([k for k, _ in short_words])
        appears = {}

        a = []
        for w in words:
            if w in keywords:
                a.append(w)
            elif a:
                self._commit_continuous_words(a, appears)
                a = []
        if a:
            self._commit_continuous_words(a, appears)

        long_words = []
        for word, count in appears.items():
            # FIXME: 合理计算阈值
            if count >= 2:
                long_words.append((word, count))
        return long_words

    def _commit_continuous_words(self, continuous_words: list, appears: dict):
        end_pos = len(continuous_words)
        if end_pos > 0:
            for i in range(0, end_pos - 1):
                for j in range(i + 2, end_pos + 1):
                    word = ''.join(continuous_words[i:j])
                    if word in appears:
                        appears[word] += 1
                    else:
                        appears[word] = 1

    def _extract_short_keywords(self, words) -> list:
        freq = {}
        for w in words:
            if self.stopwords.is_stopword(w):
                continue
            if w in freq:
                freq[w] += 1
            else:
                freq[w] = 1

        tfidf = {}
        for k in freq:
            tfidf[k] = freq[k] * self.idf.word_idf(k)

        t = []
        for w in tfidf:
            t.append((w, tfidf[w]))
        t.sort(key=lambda x: x[1], reverse=True)

        topk = 500
        # FIXME: 根据文本长度等信息动态调整简单关键词的数量
        result = []
        for w, _ in t:
            if freq[w] > 1:
                result.append((w, freq[w]))
            if len(result) >= topk:
                break
        return result

    def _rerank_words(self, freq: Dict[str, int]) -> list:
        res = []
        for k in freq:
            s = math.log2(freq[k] + 0.5) * (freq[k] / (freq[k] + self.dict.word_freq(k)))
            # FIXME: 融合词汇本身的信息量
            res.append((k, s))
        res.sort(key=lambda x: x[1], reverse=True)
        return res


_keywords_extractor = KeywordsExtractor()


def extract_keywords(doc, with_weight=False):
    # 繁体转简体
    doc = traditional_to_simple(doc)
    # 过滤无效字符
    doc = filter_invalid_chars(doc)
    keywords = _keywords_extractor.extract_keywords(doc)
    if not with_weight:
        keywords = [k[0] for k in keywords]
    return keywords



import os
from collections import defaultdict
from os import listdir
from os.path import join, dirname, isdir


class DictLoader:
    def __init__(self):
        self.dict = self._load_dict()

    def word_freq(self, word):
        return self.dict[word]

    @staticmethod
    def _load_dict():
        res = defaultdict(lambda: 0)
        d = join(dirname(__file__), 'data', 'dicts')
        for name in os.listdir(d):
            if name.endswith('.txt'):
                dict_file = join(d, name)
                with open(dict_file) as f:
                    for line in f:
                        s = line.split()
                        res[s[0]] = max(res[s[0]], int(s[1]))
        return res


class IDFLoader:
    def __init__(self):
        self.idf = self._load_idf()
        self.median_idf = sorted(self.idf.values())[len(self.idf) // 2]

    def word_idf(self, word):
        return self.idf.get(word, self.median_idf)

    @staticmethod
    def _load_idf():
        idf_file = join(dirname(__file__), 'data', 'idf.txt')
        res = {}
        with open(idf_file) as f:
            while True:
                line = f.readline()
                if not line:
                    break
                s = line.split()
                res[s[0]] = float(s[1])
        return res


class StopwordsLoader:
    def __init__(self):
        self.stopwords = self._load_stopwords()

    def is_stopword(self, word):
        return word in self.stopwords

    def remove(self, word):
        self.stopwords.remove(word)

    @staticmethod
    def _load_stopwords():
        file = join(dirname(__file__), 'data', 'stopwords')
        res = set()
        files = []
        if isdir(file):
            for fname in listdir(file):
                if fname.endswith('.txt'):
                    files.append(join(file, fname))
        else:
            files.append(file)
        for fname in files:
            with open(fname) as f:
                while True:
                    line = f.readline()
                    if not line:
                        break
                    s = line.strip()
                    res.add(s)
        return res



import re
from multiprocessing import Lock
from os.path import join, dirname

_zh_hans = {}
_zh_hans_lock = Lock()


def _load_zh_hans():
    file = join(dirname(__file__), 'data', 'zh_hans.txt')
    with open(file) as f:
        while True:
            line = f.readline()
            if not line:
                break
            s = line.split()
            _zh_hans[ord(s[0])] = ord(s[1])


def traditional_to_simple(text: str) -> str:
    with _zh_hans_lock:
        if len(_zh_hans) == 0:
            _load_zh_hans()

    return text.translate(_zh_hans)


_invalid_chars_reg = re.compile(
    r'[^' +
    r'a-zA-Z0-9' +
    r',.!?()@&_\-' +
    r'\u4e00-\u9fa5' +
    r'\u3002\uff1b\uff0c\uff1a\u201c\u201d\uff08\uff09\u3001\uff1f\u300a\u300b' +
    r']'
)


def filter_invalid_chars(text: str, char: str = None) -> str:
    if not char:
        char = ' '
    return _invalid_chars_reg.sub(char, text)



from abc import abstractmethod, ABCMeta


class WordFilter(metaclass=ABCMeta):
    @abstractmethod
    def filter(self, word: str) -> bool:
        pass


class WordFilterChain(WordFilter):
    def __init__(self, *word_filters: WordFilter):
        super().__init__()
        self.word_filters = list(word_filters)

    def filter(self, word: str) -> bool:
        for f in self.word_filters:
            if f.filter(word):
                return True
        return False


class NumericFilter(WordFilter):
    NUMERIC = set('0123456789〇一二三四五六七八九十百千万亿元角分')

    def filter(self, word: str):
        if self._is_number(word):
            return True
        nc = 0
        for i in word:
            if i in self.NUMERIC:
                nc += 1
        if nc * 2 >= len(word):
            return True
        return False

    def _is_number(self, v):
        if v:
            v = v.strip('%')
        try:
            v = float(v)
        except ValueError:
            return False
        return True
